import asyncio
import logging
import json

from aiokafka import AIOKafkaProducer

from pipeline import settings


log = logging.getLogger(__name__)


class Producer():
    def __init__(self):
        self.hosts = settings.PRODUCER_HOSTS
        self.topic = settings.PRODUCER_TOPIC
        self.producer = None
        self.connected = False
        self.loop = asyncio.get_event_loop()

    async def connect(self) -> None:
        raise NotImplemented("No driver specified")

    async def disconnect(self) -> None:
        raise NotImplemented("No driver specified")

    async def produce_message(self, message: dict, topic: str=None) -> None:
        if not topic:
            topic = settings.PRODUCER_TOPIC


class MockProducer(Producer):
    async def connect(self) -> None:
        log.debug('Connected to mock producer')
        self.connected = True

    async def disconnect(self) -> None:
        self.connected = False

    async def produce_message(self, message: dict, topic: str=None) -> None:
        log.debug('Mocking a message on topic {topic}: {message}'.format(
            topic=topic,
            message=message,
        ))


class KafkaProducer(Producer):
    async def connect(self) -> None:
        log.debug('Connecting to Kafka stream.')
        self.producer = AIOKafkaProducer(
            loop=self.loop,
            bootstrap_servers=self.hosts
        )
        await self.producer.start()
        self.connected = True

    async def disconnect(self) -> None:
        await self.producer.stop()
        self.connected = False

    async def produce_message(self, message: dict, topic: str=None) -> None:
        log.debug('Trying to produce message: {msg}'.format(msg=message))

        if not self.connected:
            log.debug('Producer not connected')
            return

        await super().produce_message(message=message, topic=topic)
        await self.producer.send_and_wait(topic, json.dumps(message).encode())


def get_producer() -> Producer:
    driver = settings.PRODUCER_DRIVER
    
    if driver == 'kafka':
        return KafkaProducer()
    else:
        return MockProducer()




import logging
import json
import asyncio
from typing import (
    Optional,
    List,
    Callable,
    Tuple,
)

from gmqtt import Client, Subscription
from gmqtt.mqtt.constants import MQTTv311

from pipeline.producers import get_producer


log = logging.getLogger(__name__)


def _on_connect(client, flags, rc, properties):
    log.debug('{client_id}: CONNECTED'.format(client_id=client._client_id))

    
def _on_message(client, topic, payload, qos, properties):
    log.debug('RECV: {payload}'.format(payload=payload))

    
def _on_disconnect(client, packet, exc=None):
    log.debug('{client_id}: DISCONNECTED'.format(client_id=client._client_id))

    
def _on_subscribe(client, mid, qos):
    log.debug('{client_id}: SUBSCRIBED'.format(client_id=client._client_id))


class MQTTClient(Client):
    def __init__(
            self,
            username: str=None,
            password: str=None,
            connected: bool=False,
            on_connect: Callable=None,
            on_message: Callable=None,
            on_disconnect: Callable=None,
            on_subscribe: Callable=None,
    ) -> None:        
        super().__init__(None)

        if not username is None:
            self.set_auth_credentials(username, password)
        
        self.producer = get_producer()
        self.connected = connected
        
        self.on_connect = on_connect or _on_connect
        self.on_message = on_message or _on_message
        self.on_disconnect = on_disconnect or _on_disconnect
        self.on_subscribe = on_subscribe or _on_subscribe

    async def pipe_message(self, message: str, topic: str) -> None:
        await self.producer.produce_message(
            message=message,
            topic=topic,
        )

    async def connect(self, uri: str, topics: Tuple[str, int]) -> None:
        log.debug('Connecting to MQTT broker.')        
        try:
            await super().connect(uri, 1883, keepalive=60, version=MQTTv311)
            await self.producer.connect()

            subscriptions = [Subscription(t[0], qos=t[1]) for t in topics]
            self.subscribe(subscriptions, subscription_identifier=1)            
            self.connected = True
        except:
            self.connected = False

            
async def serialize(data):
    data = data.decode('utf-8')
    try:
        return json.loads(data)
    except json.decoder.JSONDecodeError:
        return data



# system config
LOG_DIR = None

# LoRa server settings
LORA_MQTT_TOPICS = None
LORA_MQTT_URI = None
LORA_DRIVER = None
LORA_MQTT_USERNAME = None
LORA_MQTT_PASSWORD = None

# Producer settings
PRODUCER_DRIVER = None
PRODUCER_TOPIC = None
PRODUCER_HOSTS = None



from typing import Dict
from pathlib import Path

import yaml
from yaml.parser import ParserError

from pipeline.types import T


class ConfigLoadError(Exception):
    pass


def load_config_file(config_file: Path) -> Dict[str, T]:
    """
    Load a YAML config file from the given path.
    Raises:
        ConfigLoadError if config file could not be loaded.
    Returns:
         The parsed config values.
    """

    try:
        file = open(str(config_file), 'r')
        config = yaml.load(file)
        file.close()

        if type(config) is not dict:
            raise TypeError('Config must be a map at root level.')

    except IOError as e:
        raise ConfigLoadError(
            'Failed opening config file: {msg}.'.format(msg=e)
        ) from e

    except (ParserError, TypeError) as e:
        raise ConfigLoadError(
            'Failed parsing config file: {msg}'.format(msg=e)
        ) from e

    else:
        return config



import logging
import asyncio
import functools
import signal
from typing import Coroutine

from pipeline import STOP


log = logging.getLogger(__name__)


async def task_wrapper(task: Coroutine):
    try:
        await task(stop_signal=STOP)
    except asyncio.CancelledError:
        log.debug('Cancelled: {task}'.format(task=task.__name__))
    except asyncio.TimeoutError:
        log.debug('Timed out: {task}. Restarting.'.format(task=task.__name__))

        
def ask_exit(*args):
    STOP.set()

    
def main(task):
    loop = asyncio.get_event_loop()

    # connect signal handler for stopping program
    stop_signals = [signal.SIGINT, signal.SIGHUP]
    for value in stop_signals:
        loop.add_signal_handler(value, ask_exit)

    log.debug('Starting loop.')
    loop.run_until_complete(asyncio.gather(
        task_wrapper(task),
    ))

    log.debug('Main finished.')



import sys
import logging

import pipeline
from pipeline import settings, init_logging
from pipeline.config import load_config_file, ConfigLoadError
from pipeline.cli_args import (
    Arg,
    parse_args,
    ValidationError
)


args = [
    Arg(
        name='config_file',
        data_type='path',
        ensure_exists='file',
        required=True,
    ),
]


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def cli(task) -> None:
    """
    Initialize args, config, logging, settings, and runs main function.
    Use sys.exit to terminate program.
    """

    # parse cli args

    try:
        parsed_args = parse_args(args)
    except ValidationError as e:
        eprint('Argument error: {msg}'.format(msg=e))
        sys.exit(1)

    # load config file
    try:
        config_values = load_config_file(
            config_file=parsed_args.get('config_file'),
        )

        # set settings from config_values
        for name, value in config_values.items():
            if hasattr(settings, name):
                setattr(settings, name, value)

    except ConfigLoadError as e:
        eprint('No config file loaded: {msg}, going for default. God bless.'.format(msg=e))

    # init logging
    init_logging()

    # get logger for this file
    log = logging.getLogger(__name__)

    # log program info
    log.info('Trell Data Pipeline commencing: version={version}'.format(
        version=pipeline.__version__
    ))

    # run main program
    from pipeline.main import main
    main(task)
    # main(task, args)
    sys.exit(0)



__version__ = '0.2.2'

import asyncio
from typing import Optional
import logging
import logging.handlers
from pathlib import Path
import sys
from datetime import datetime

from pipeline import settings


STOP = asyncio.Event()


class DebugOnlyFilter(logging.Filter):
    def filter(self, record):
        return record.levelno == logging.DEBUG


class FileFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        created = datetime.utcfromtimestamp(record.created)
        return created.strftime('%Y-%m-%d %H:%M:%S:%f')


def init_logging() -> None:
    console_formatter = logging.Formatter(
        fmt='{asctime:s} {name:42s} {levelname:8s} {message:s}',
        style='{',
    )
    file_formatter = FileFormatter(
        fmt='{asctime:s} {name:s} {levelname:s} {message:s}',
        style='{',
    )

    # root logger
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    # console output
    stdout = logging.StreamHandler(stream=sys.stdout)
    stdout.setFormatter(console_formatter)
    stdout.setLevel(logging.DEBUG)
    root.addHandler(stdout)

    # log files
    if settings.LOG_DIR is not None:
        log_dir = Path(settings.LOG_DIR)

        # debug file
        debug_file = logging.handlers.RotatingFileHandler(
            filename=log_dir / 'trell-data-pipeline.debug.log',
            maxBytes=5 * 1000 * 1000,
            backupCount=2,
        )
        debug_file.setFormatter(file_formatter)
        debug_file.setLevel(logging.DEBUG)
        root.addHandler(debug_file)

        # info file
        info_file = logging.handlers.RotatingFileHandler(
            filename=log_dir / 'trell_lora_pipeline.info.log',
            maxBytes=5 * 1000 * 1000,
            backupCount=2,
        )
        info_file.setFormatter(file_formatter)
        info_file.setLevel(logging.INFO)
        root.addHandler(info_file)

    # disable asyncio debug log
    asyncio_logger = logging.getLogger('asyncio')
    asyncio_logger.setLevel(logging.WARNING)

    # disable gmqtt debug log
    gmqtt_logger = logging.getLogger('gmqtt')
    gmqtt_logger.setLevel(logging.WARNING)

    # disable transitions debug log
    transitions_logger = logging.getLogger('transitions')
    transitions_logger.setLevel(logging.WARNING)

    # disable aiokafka debug log
    aiokafka_logger = logging.getLogger('aiokafka')
    aiokafka_logger.setLevel(logging.WARNING)



from typing import Optional, List, Dict
import argparse
import os
from pathlib import Path

from pipeline.types import T


class ValidationError(Exception):
    pass


class Arg:
    def __init__(
        self,
        name: str,
        data_type: str,
        ensure_exists: Optional[bool]=False,
        argparse_name: Optional[str]=None,
        argparse_dest: Optional[str]=None,
        argparse_action: Optional[str]=None,
        envvar: Optional[str]=None,
        required: bool=False,
        default: Optional[T]=None,
    ):
        self.name = name
        self.data_type = data_type
        self.ensure_exists = ensure_exists

        if argparse_name is not None:
            self.argparse_name = argparse_name
        else:
            self.argparse_name = '--{name}'.format(name=name.replace('_', '-'))

        if argparse_dest is not None:
            self.argparse_dest = argparse_dest
        else:
            self.argparse_dest = name

        self.argparse_action = argparse_action
        if argparse_action is None:
            if data_type == 'bool':
                self.argparse_action = 'store_true'

        if envvar is not None:
            self.envvar = envvar
        else:
            self.envvar = 'TLP_{name}'.format(name=name.upper())

        self.required = required
        self.default = default

    def __str__(self):
        return (
            'Arg('
            'name={self.name!r}, '
            'data_type={self.data_type!r}, '
            'argparse_name={self.argparse_name!r}, '
            'argparse_dest={self.argparse_dest!r}, '
            'argparse_action={self.argparse_action!r}, '
            'envvar={self.envvar!r}, '
            'required={self.required!r}, '
            'default={self.default!r}'
            ')'
        ).format(self=self)


def parse_args(args: List[Arg]) -> Dict[str, T]:
    results = {}

    # use argparse to parse cli args
    parser = argparse.ArgumentParser()

    for arg in args:
        parser_kwargs = dict(
            dest=arg.argparse_dest,
        )

        if arg.argparse_action is not None:
            parser_kwargs['action'] = arg.argparse_action

        parser.add_argument(arg.argparse_name, **parser_kwargs)

    parsed_args = parser.parse_args()

    # after argparse, extend with additional functionality
    for arg in args:

        # read value from either cli arg or env var
        arg_value = getattr(parsed_args, arg.argparse_dest, None)
        env_value = None

        if arg.envvar is not None:
            if arg.envvar in os.environ:
                env_value = os.environ[arg.envvar]

        value = arg_value or env_value

        # if value is not present, and default is available, use default
        if value is None and arg.default is not None:
            value = arg.default

        # if value is required, and value not present, raise error
        if arg.required and value is None:
            raise ValidationError('Argument {name} is required'.format(
                name=arg.name
            ))

        # if there is a value, parse it according to data type
        if value is not None:
            if arg.data_type == 'str':
                value = str(value)

            if arg.data_type == 'int':
                value = int(value)

            if arg.data_type == 'bool':
                value = value in ['true', 'True', 'TRUE', '1', 'yes']

            if arg.data_type == 'path':
                value = Path(value)

        # if data_type is path and ensure_exists is true, ensure path exists
        if value is not None and arg.data_type == 'path' and arg.ensure_exists:
            if any([
                arg.ensure_exists == 'file' and not value.is_file(),
                arg.ensure_exists == 'dir' and not value.is_dir(),
            ]):
                raise ValidationError(
                    'Path does not exist: {name}={value}'.format(
                        name=arg.name,
                        value=value,
                    )
                )

        results[arg.name] = value

    return results



#!/usr/bin/env python
from setuptools import find_packages, setup
from pipeline import __version__

setup(
    name='trell-data-pipeline',
    version=__version__,
    description='Pipes various ingress streams to another egress stream. I.e. MQTT -> Kafka',
    packages=find_packages(exclude=['test', 'test.*']),
    url = 'https://github.com/frisellcpl/trell-data-pipeline',
    author = 'Johan Frisell',
    author_email = 'johan@trell.se',
    include_package_data=True,
    install_requires=[
        'aiokafka==0.5.2',
        'PyYAML==5.3',
        'gmqtt==0.5.5',
    ],
    classifiers = [
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    extras_require={
        'dev': [
            'pytest==2.9.2',
        ],
    },
    zip_safe=True
)



import base64
import asyncio
import logging

from abc import ABC

log = logging.getLogger(__name__)


INDOOR_SENSOR = 'indoor_sensor'
DOOR_LOCK_SENSOR = 'door_lock_sensor'


class Decoder(ABC):
    def __init__(self, data):
        self.raw = data
        self.data = None
        
    async def decode(self):
        return {k: v for k,v in self.raw}

    async def is_valid(self):
        self.data = await self.decode()            

        if self.data:
            return True
            
        return False

    @property
    def sensor_type(self):
        pass

    
class SensitiviaDecoder(Decoder):
    key_map = {
        'Battery': 'b',
        'AvgTemperature': 't',
        'Humidity': 'h',
        'Lux': 'l',
        'fcnt': 'fcnt',
        'rssi': 'rssi',
        'port': 'port',
    }

    async def decode(self):
        data = {}
        
        for k,v in self.raw:
            try:
                data[self.key_map[k]] = v
            except KeyError:
                continue
            
        return data

    async def is_valid(self):
        return await super().is_valid()
    
    @property
    def sensor_type(self):
        return INDOOR_SENSOR


class DoorLockDecoder(Decoder):
    async def decode(self):
        data = super().decode()

        o = data.get('DI1', None)
        l = data.get('DI2', None)

        return {
            'open': None if o is None else o == 0,
            'locked': None if l is None else l == 0,
        }

    async def is_valid(self):
        return await super().is_valid()
    
    @property
    def sensor_type(self):
        return DOOR_LOCK_SENSOR


class ElsysDecoder(Decoder):
    TYPE_TEMP = 0x01  # Temp 2 bytes -3276. -->
    TYPE_RH = 0x02  # Humidity 1 byte  0-100%
    TYPE_ACC = 0x03  # Acceleration 3 bytes X,Y,Z -128 --> 127 +/-63=1G
    TYPE_LIGHT = 0x04  # Light 2 bytes 0-->65535 Lux
    TYPE_MOTION = 0x05  # No of motion 1 byte  0-255
    TYPE_CO2 = 0x06  # Co2 2 bytes 0-65535 ppm
    TYPE_VDD = 0x07  # VDD 2byte 0-65535mV
    TYPE_ANALOG1 = 0x08  # VDD 2byte 0-65535mV
    TYPE_GPS = 0x09  # 3bytes lat 3bytes long binary
    TYPE_PULSE1 = 0x0A  # 2bytes relative pulse count
    TYPE_PULSE1_ABS = 0x0B  # 4bytes no 0->0xFFFFFFFF
    TYPE_EXT_TEMP1 = 0x0C  # 2bytes -3276.5C-->3276.5C
    TYPE_EXT_DIGITAL = 0x0D  # 1bytes value 1 or 0
    TYPE_EXT_DISTANCE = 0x0E  # 2bytes distance in mm
    TYPE_ACC_MOTION = 0x0F  # 1byte number of vibration/motion
    TYPE_IR_TEMP = 0x10  # 2bytes internal temp 2bytes external temp -3276.5C-->3276.5C
    TYPE_OCCUPANCY = 0x11  # 1byte data
    TYPE_WATERLEAK = 0x12  # 1byte data 0-255
    TYPE_GRIDEYE = 0x13  # 65byte temperature data 1byte ref+64byte external temp
    TYPE_PRESSURE = 0x14  # 4byte pressure data (hPa)
    TYPE_SOUND = 0x15  # 2byte sound data (peak/avg)
    TYPE_PULSE2 = 0x16  # 2bytes 0-->0xFFFF
    TYPE_PULSE2_ABS = 0x17  # 4bytes no 0->0xFFFFFFFF
    TYPE_ANALOG2 = 0x18  # 2bytes voltage in mV
    TYPE_EXT_TEMP2 = 0x19
                    
    async def is_valid(self):
        try:            
            self.data = await self.decode()            

            if self.data:
                return True
            
        except IndexError:
            return False # TODO: reraise a better error message and handle in caller.

        return False
    
    @property
    def sensor_type(self):
        return INDOOR_SENSOR

    async def _decode_hex(self, data):
        obj = {}
        i = 0
        while i < len(data):
            if data[i] == self.TYPE_TEMP:
                num = int(data[i+1] << 8 | data[i+2])
                if num > 0x7FFF:
                    num -= 0x10000
                obj['t'] = num/10
                i += 2
            elif data[i] == self.TYPE_RH:
                obj['h'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_ACC:
                i += 2
            elif data[i] == self.TYPE_LIGHT:
                obj['l'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_MOTION:
                obj['m'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_CO2:
                obj['c'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_VDD:
                obj['b'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_ANALOG1:
                obj['a1'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_GPS:
                i += 6
            elif data[i] == self.TYPE_PULSE1:
                obj['p1'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_PULSE1_ABS:
                i += 4
            elif data[i] == self.TYPE_EXT_TEMP1:
                i += 2
            elif data[i] == self.TYPE_EXT_DIGITAL:
                obj['edig'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_EXT_DISTANCE:
                obj['edis'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_ACC_MOTION:
                obj['am'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_IR_TEMP:
                it = int(data[i+1] << 8 | data[i+2])
                if it > 0x7FFF:
                    it -= 0x10000

                et = int(data[i+3] << 8 | data[i+4])
                if et > 0x7FFF:
                    et -= 0x10000

                obj['irit'] = it/10
                obj['iret'] = et/10
                
                i += 4
            elif data[i] == self.TYPE_OCCUPANCY:
                obj['o'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_WATERLEAK:
                obj['w'] = data[i+1]
                i += 1
            elif data[i] == self.TYPE_GRIDEYE:
                i += 65
            elif data[i] == self.TYPE_PRESSURE:
                i += 4
            elif data[i] == self.TYPE_SOUND:
                i += 2
            elif data[i] == self.TYPE_PULSE2:
                obj['p2'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_PULSE2_ABS:
                i += 4
            elif data[i] == self.TYPE_ANALOG2:
                obj['a2'] = int(data[i+1] << 8 | data[i+2])
                i += 2
            elif data[i] == self.TYPE_EXT_TEMP2:
                i += 2
            else:
                log.error('Couldnt decode hex, at pos {i} value {v}'.format(
                    i=i,
                    v=data[i],
                ))
                i = len(data)
            i = i+1
            
        return obj


class AdvantechElsysDecoder(ElsysDecoder):        
    async def decode(self):
        data = {}
        
        for k,v in self.raw:
            if k == 'hex':
                data.update(await self._decode_hex(
                    data=bytes.fromhex(v)
                ))
            else:
                data[k] = v

        log.debug('Decoded {d} from {r}'.format(d=data, r=self.raw))
        return data


class ChirpElsysDecoder(ElsysDecoder):
    async def decode(self):
        data = await self._decode_hex(
            data=base64.b64decode(self.raw)
        )
        
        log.debug('Decoded {d} from {r}'.format(d=data, r=self.raw))
        return data
